from pathlib import Path

from ragraph import plot
from ragraph.graph import Graph
from ragraph.io.grip import from_grip


def test_grip_format(datadir: Path, tmpdir: Path):
    g: Graph = from_grip(datadir / "grip" / "grip.xml")
    assert g.nodes, "I want nodes."

    fig = plot.mdm(
        g.leafs, g.edges, style=plot.Style(piemap=(dict(display="kinds", mode="relative")))
    )
    fig.write_image(tmpdir / "grip.svg", format="svg")


def test_grip_objecttype(datadir: Path, tmpdir: Path):
    g: Graph = from_grip(datadir / "grip" / "lock-head-example.xml")

    assert g.nodes, "I want nodes."

    objecttypes = sorted([n.name for n in g.nodes if n.kind == "objecttype"])

    assert objecttypes == [
        "LBS - S_L1.3.1 - Nivelleersysteem constructief | OBT-00062",
        "LBS - S_L1.4.1 - Sluishoofd | OBT-00060",
        "LBS - S_L1.4.1.5 - IJsbestrijdingsysteem | OBT-00065",
        "LBS - S_L1.4.3 - Sluisdeur (constructief) | OBT-00058",
        "LBS - S_L1.4.4 - Dynamisch sluisdeur systeem | OBT-00057",
        "LBS - S_L5.1 - Nivelleersysteem | OBT-00061",
        "LBS - S_L6.1.2 - Bewegingswerk nivelleerschuif (constructief) | OBT-00064",
        "LBS - S_L6.1.2 - Bewegingswerk sluisdeur (constructief) | OBT-00063",
        "NEN2767 - 101 - Aandrijving en bewegingswerk (elektrohydraulisch) | OBT-00067",
        "NEN2767 - 102 - Aandrijving en bewegingswerk (elektromechanisch) | OBT-00068",
        "NEN2767 - 1203 - Frame | OBT-00066",
        "NEN2767 - 1375 - Nivelleerschuif | OBT-00069",
        "NEN2767 - 1383 - Omloopriool | OBT-00070",
        "NEN2767 - 198 - Sluisdeur (hef, punt, rol) | OBT-00072",
        "NEN2767 - 199 - Sluishoofd | OBT-00071",
        "UCS - Beschoeiing - Sluisdeur - Punt - Hout | OBT-00024",
        "UCS - Beschoeiing - Sluisdeur - Punt - Staal | OBT-00025",
        "UCS - Beschoeiing - Sluisdeur - Punt | OBT-00031",
        "UCS - Beschoeiing - Sluisdeur | OBT-00030",
        "UCS - Beschoeiing | OBT-00029",
        "UCS - Bewegingswerk - Nivelleersysteem -  Elektrohydraulische cilinder | OBT-00051",
        "UCS - Bewegingswerk - Nivelleersysteem - Elektromechanische cilinder | OBT-00052",
        "UCS - Bewegingswerk - Nivelleersysteem | OBT-00046",
        "UCS - Bewegingswerk - Sluisdeur - Elektrohydraulische cilinder | OBT-00042",
        "UCS - Bewegingswerk - Sluisdeur - Elektromechanische cilinder | OBT-00041",
        "UCS - Bewegingswerk - Sluisdeur - Horizontaal lierwerk | OBT-00043",
        "UCS - Bewegingswerk - Sluisdeur - Vertikaal  lierwerk | OBT-00044",
        "UCS - Bewegingswerk - Sluisdeur | OBT-00045",
        "UCS - Bewegingswerk | OBT-00040",
        "UCS - Frame - Sluisdeur - Punt - Hout | OBT-00049",
        "UCS - Frame - Sluisdeur - Punt - Staal | OBT-00050",
        "UCS - Frame - Sluisdeur - Punt | OBT-00048",
        "UCS - Frame - Sluisdeur | OBT-00047",
        "UCS - Frame | OBT-00033",
        "UCS - IJsbestrijdingsinstallatie - Bellenscherm | OBT-00054",
        "UCS - IJsbestrijdingsinstallatie | OBT-00053",
        "UCS - Nivelleersysteem - Omloopriolen bodemvulling | OBT-00038",
        "UCS - Nivelleersysteem - Omloopriolen met woelkelder | OBT-00037",
        "UCS - Nivelleersysteem - Omloopriolen overdwars | OBT-00036",
        "UCS - Nivelleersysteem - Omloopriolen wandvulling | OBT-00039",
        "UCS - Nivelleersysteem - Rinketten | OBT-00035",
        "UCS - Nivelleersysteem | OBT-00034",
        "UCS - Sluisdeur - Hef | OBT-00021",
        "UCS - Sluisdeur - Punt | OBT-00020",
        "UCS - Sluisdeur - Rol | OBT-00022",
        "UCS - Sluisdeur - Vierkantkerend | OBT-00023",
        "UCS - Sluisdeur | OBT-00028",
        "UCS - Sluishoofd - standaard | OBT-00059",
    ]

    inheritance_edges = [(e.source.name, e.target.name) for e in g.edges if e.kind == "inheritance"]

    assert len(inheritance_edges) == len(
        [
            (
                "UCS - Nivelleersysteem | OBT-00034",
                "LBS - S_L1.3.1 - Nivelleersysteem constructief | OBT-00062",
            ),
            ("UCS - Sluishoofd - standaard | OBT-00059", "LBS - S_L1.4.1 - Sluishoofd | OBT-00060"),
            (
                "UCS - IJsbestrijdingsinstallatie | OBT-00053",
                "LBS - S_L1.4.1.5 - IJsbestrijdingsysteem | OBT-00065",
            ),
            (
                "UCS - Sluisdeur | OBT-00028",
                "LBS - S_L1.4.3 - Sluisdeur (constructief) | OBT-00058",
            ),
            (
                "UCS - Sluisdeur | OBT-00028",
                "LBS - S_L1.4.4 - Dynamisch sluisdeur systeem | OBT-00057",
            ),
            ("UCS - Nivelleersysteem | OBT-00034", "LBS - S_L5.1 - Nivelleersysteem | OBT-00061"),
            (
                "UCS - Bewegingswerk - Nivelleersysteem | OBT-00046",
                "LBS - S_L6.1.2 - Bewegingswerk nivelleerschuif (constructief) | OBT-00064",
            ),
            (
                "UCS - Bewegingswerk - Sluisdeur | OBT-00045",
                "LBS - S_L6.1.2 - Bewegingswerk sluisdeur (constructief) | OBT-00063",
            ),
            (
                "UCS - Bewegingswerk - Nivelleersysteem -  "
                "Elektrohydraulische cilinder | OBT-00051",
                "NEN2767 - 101 - Aandrijving en bewegingswerk (elektrohydraulisch) | OBT-00067",
            ),
            (
                "UCS - Bewegingswerk - Sluisdeur - Elektrohydraulische cilinder | OBT-00042",
                "NEN2767 - 101 - Aandrijving en bewegingswerk (elektrohydraulisch) | OBT-00067",
            ),
            (
                "UCS - Bewegingswerk - Nivelleersysteem - Elektromechanische cilinder | OBT-00052",
                "NEN2767 - 102 - Aandrijving en bewegingswerk (elektromechanisch) | OBT-00068",
            ),
            (
                "UCS - Bewegingswerk - Sluisdeur - Elektromechanische cilinder | OBT-00041",
                "NEN2767 - 102 - Aandrijving en bewegingswerk (elektromechanisch) | OBT-00068",
            ),
            (
                "UCS - Bewegingswerk - Sluisdeur - Horizontaal lierwerk | OBT-00043",
                "NEN2767 - 102 - Aandrijving en bewegingswerk (elektromechanisch) | OBT-00068",
            ),
            (
                "UCS - Bewegingswerk - Sluisdeur - Vertikaal  lierwerk | OBT-00044",
                "NEN2767 - 102 - Aandrijving en bewegingswerk (elektromechanisch) | OBT-00068",
            ),
            ("UCS - Frame | OBT-00033", "NEN2767 - 1203 - Frame | OBT-00066"),
            (
                "UCS - Nivelleersysteem - Rinketten | OBT-00035",
                "NEN2767 - 1375 - Nivelleerschuif | OBT-00069",
            ),
            (
                "UCS - Nivelleersysteem - Omloopriolen bodemvulling | OBT-00038",
                "NEN2767 - 1383 - Omloopriool | OBT-00070",
            ),
            (
                "UCS - Nivelleersysteem - Omloopriolen met woelkelder | OBT-00037",
                "NEN2767 - 1383 - Omloopriool | OBT-00070",
            ),
            (
                "UCS - Nivelleersysteem - Omloopriolen overdwars | OBT-00036",
                "NEN2767 - 1383 - Omloopriool | OBT-00070",
            ),
            (
                "UCS - Nivelleersysteem - Omloopriolen wandvulling | OBT-00039",
                "NEN2767 - 1383 - Omloopriool | OBT-00070",
            ),
            (
                "UCS - Sluisdeur | OBT-00028",
                "NEN2767 - 198 - Sluisdeur (hef, punt, rol) | OBT-00072",
            ),
            ("UCS - Sluishoofd - standaard | OBT-00059", "NEN2767 - 199 - Sluishoofd | OBT-00071"),
            ("UCS - Beschoeiing - Sluisdeur | OBT-00030", "UCS - Beschoeiing | OBT-00029"),
            (
                "UCS - Beschoeiing - Sluisdeur - Punt | OBT-00031",
                "UCS - Beschoeiing - Sluisdeur | OBT-00030",
            ),
            (
                "UCS - Beschoeiing - Sluisdeur - Punt - Hout | OBT-00024",
                "UCS - Beschoeiing - Sluisdeur - Punt | OBT-00031",
            ),
            (
                "UCS - Beschoeiing - Sluisdeur - Punt - Staal | OBT-00025",
                "UCS - Beschoeiing - Sluisdeur - Punt | OBT-00031",
            ),
            ("UCS - Bewegingswerk - Sluisdeur | OBT-00045", "UCS - Bewegingswerk | OBT-00040"),
            (
                "UCS - Bewegingswerk - Nivelleersysteem | OBT-00046",
                "UCS - Bewegingswerk | OBT-00040",
            ),
            (
                "UCS - Bewegingswerk - Nivelleersysteem -  "
                "Elektrohydraulische cilinder | OBT-00051",
                "UCS - Bewegingswerk - Nivelleersysteem | OBT-00046",
            ),
            (
                "UCS - Bewegingswerk - Nivelleersysteem - "
                "Elektromechanische cilinder | OBT-00052",
                "UCS - Bewegingswerk - Nivelleersysteem | OBT-00046",
            ),
            (
                "UCS - Bewegingswerk - Sluisdeur - Vertikaal  lierwerk | OBT-00044",
                "UCS - Bewegingswerk - Sluisdeur | OBT-00045",
            ),
            (
                "UCS - Bewegingswerk - Sluisdeur - Horizontaal lierwerk | OBT-00043",
                "UCS - Bewegingswerk - Sluisdeur | OBT-00045",
            ),
            (
                "UCS - Bewegingswerk - Sluisdeur - Elektrohydraulische cilinder | OBT-00042",
                "UCS - Bewegingswerk - Sluisdeur | OBT-00045",
            ),
            (
                "UCS - Bewegingswerk - Sluisdeur - Elektromechanische cilinder | OBT-00041",
                "UCS - Bewegingswerk - Sluisdeur | OBT-00045",
            ),
            ("UCS - Frame - Sluisdeur | OBT-00047", "UCS - Frame | OBT-00033"),
            ("UCS - Frame - Sluisdeur - Punt | OBT-00048", "UCS - Frame - Sluisdeur | OBT-00047"),
            (
                "UCS - Frame - Sluisdeur - Punt - Hout | OBT-00049",
                "UCS - Frame - Sluisdeur - Punt | OBT-00048",
            ),
            (
                "UCS - Frame - Sluisdeur - Punt - Staal | OBT-00050",
                "UCS - Frame - Sluisdeur - Punt | OBT-00048",
            ),
            (
                "UCS - IJsbestrijdingsinstallatie - Bellenscherm | OBT-00054",
                "UCS - IJsbestrijdingsinstallatie | OBT-00053",
            ),
            (
                "UCS - Nivelleersysteem - Rinketten | OBT-00035",
                "UCS - Nivelleersysteem | OBT-00034",
            ),
            (
                "UCS - Nivelleersysteem - Omloopriolen overdwars | OBT-00036",
                "UCS - Nivelleersysteem | OBT-00034",
            ),
            (
                "UCS - Nivelleersysteem - Omloopriolen met woelkelder | OBT-00037",
                "UCS - Nivelleersysteem | OBT-00034",
            ),
            (
                "UCS - Nivelleersysteem - Omloopriolen bodemvulling | OBT-00038",
                "UCS - Nivelleersysteem | OBT-00034",
            ),
            (
                "UCS - Nivelleersysteem - Omloopriolen wandvulling | OBT-00039",
                "UCS - Nivelleersysteem | OBT-00034",
            ),
            ("UCS - Sluisdeur - Vierkantkerend | OBT-00023", "UCS - Sluisdeur | OBT-00028"),
            ("UCS - Sluisdeur - Rol | OBT-00022", "UCS - Sluisdeur | OBT-00028"),
            ("UCS - Sluisdeur - Hef | OBT-00021", "UCS - Sluisdeur | OBT-00028"),
            ("UCS - Sluisdeur - Punt | OBT-00020", "UCS - Sluisdeur | OBT-00028"),
            (
                "VARIANT - Bellenscherm | OBJ-02472",
                "UCS - IJsbestrijdingsinstallatie - Bellenscherm | OBT-00054",
            ),
            (
                "UCS - IJsbestrijdingsinstallatie - Bellenscherm | OBT-00054",
                "VARIANT - Bellenscherm | OBJ-02472",
            ),
            (
                "VARIANT - Elektrohydraulische cilinder | OBJ-02468",
                "UCS - Bewegingswerk - Sluisdeur - Elektrohydraulische cilinder | OBT-00042",
            ),
            (
                "UCS - Bewegingswerk - Sluisdeur - Elektrohydraulische cilinder | OBT-00042",
                "VARIANT - Elektrohydraulische cilinder | OBJ-02468",
            ),
            (
                "VARIANT - Elektrohydraulische cilinder | OBJ-02470",
                "UCS - Bewegingswerk - Nivelleersysteem -  "
                "Elektrohydraulische cilinder | OBT-00051",
            ),
            (
                "UCS - Bewegingswerk - Nivelleersysteem -  "
                "Elektrohydraulische cilinder | OBT-00051",
                "VARIANT - Elektrohydraulische cilinder | OBJ-02470",
            ),
            (
                "VARIANT - Elektromechanische cilinder | OBJ-02469",
                "UCS - Bewegingswerk - Sluisdeur - Elektromechanische cilinder | OBT-00041",
            ),
            (
                "UCS - Bewegingswerk - Sluisdeur - Elektromechanische cilinder | OBT-00041",
                "VARIANT - Elektromechanische cilinder | OBJ-02469",
            ),
            (
                "VARIANT - Elektromechanische cilinder | OBJ-02471",
                "UCS - Bewegingswerk - Nivelleersysteem - Elektromechanische cilinder | OBT-00052",
            ),
            (
                "UCS - Bewegingswerk - Nivelleersysteem - Elektromechanische cilinder | OBT-00052",
                "VARIANT - Elektromechanische cilinder | OBJ-02471",
            ),
            ("VARIANT - Hef | OBJ-02460", "UCS - Sluisdeur - Hef | OBT-00021"),
            ("UCS - Sluisdeur - Hef | OBT-00021", "VARIANT - Hef | OBJ-02460"),
            (
                "VARIANT - Horizontaal lierwerk | OBJ-02463",
                "UCS - Bewegingswerk - Sluisdeur - Horizontaal lierwerk | OBT-00043",
            ),
            (
                "UCS - Bewegingswerk - Sluisdeur - Horizontaal lierwerk | OBT-00043",
                "VARIANT - Horizontaal lierwerk | OBJ-02463",
            ),
            (
                "VARIANT - Hout | OBJ-02474",
                "UCS - Beschoeiing - Sluisdeur - Punt - Hout | OBT-00024",
            ),
            (
                "UCS - Beschoeiing - Sluisdeur - Punt - Hout | OBT-00024",
                "VARIANT - Hout | OBJ-02474",
            ),
            ("VARIANT - Hout | OBJ-02478", "UCS - Frame - Sluisdeur - Punt - Hout | OBT-00049"),
            ("UCS - Frame - Sluisdeur - Punt - Hout | OBT-00049", "VARIANT - Hout | OBJ-02478"),
            (
                "VARIANT - Omloopriolen bodemvulling | OBJ-06485",
                "UCS - Nivelleersysteem - Omloopriolen bodemvulling | OBT-00038",
            ),
            (
                "UCS - Nivelleersysteem - Omloopriolen bodemvulling | OBT-00038",
                "VARIANT - Omloopriolen bodemvulling | OBJ-06485",
            ),
            (
                "VARIANT - Omloopriolen met woelkelder | OBJ-02466",
                "UCS - Nivelleersysteem - Omloopriolen met woelkelder | OBT-00037",
            ),
            (
                "UCS - Nivelleersysteem - Omloopriolen met woelkelder | OBT-00037",
                "VARIANT - Omloopriolen met woelkelder | OBJ-02466",
            ),
            (
                "VARIANT - Omloopriolen overdwars | OBJ-06486",
                "UCS - Nivelleersysteem - Omloopriolen overdwars | OBT-00036",
            ),
            (
                "UCS - Nivelleersysteem - Omloopriolen overdwars | OBT-00036",
                "VARIANT - Omloopriolen overdwars | OBJ-06486",
            ),
            (
                "VARIANT - Omloopriolen wandvulling | OBJ-06484",
                "UCS - Nivelleersysteem - Omloopriolen wandvulling | OBT-00039",
            ),
            (
                "UCS - Nivelleersysteem - Omloopriolen wandvulling | OBT-00039",
                "VARIANT - Omloopriolen wandvulling | OBJ-06484",
            ),
            ("VARIANT - Punt | OBJ-02458", "UCS - Sluisdeur - Punt | OBT-00020"),
            ("UCS - Sluisdeur - Punt | OBT-00020", "VARIANT - Punt | OBJ-02458"),
            ("VARIANT - Rinketten | OBJ-02465", "UCS - Nivelleersysteem - Rinketten | OBT-00035"),
            ("UCS - Nivelleersysteem - Rinketten | OBT-00035", "VARIANT - Rinketten | OBJ-02465"),
            ("VARIANT - Rol | OBJ-02461", "UCS - Sluisdeur - Rol | OBT-00022"),
            ("UCS - Sluisdeur - Rol | OBT-00022", "VARIANT - Rol | OBJ-02461"),
            ("VARIANT - Staal | OBJ-02479", "UCS - Frame - Sluisdeur - Punt - Staal | OBT-00050"),
            ("UCS - Frame - Sluisdeur - Punt - Staal | OBT-00050", "VARIANT - Staal | OBJ-02479"),
            (
                "VARIANT - Staal | OBJ-02475",
                "UCS - Beschoeiing - Sluisdeur - Punt - Staal | OBT-00025",
            ),
            (
                "UCS - Beschoeiing - Sluisdeur - Punt - Staal | OBT-00025",
                "VARIANT - Staal | OBJ-02475",
            ),
            ("VARIANT - Standaard | OBJ-02452", "UCS - Sluishoofd - standaard | OBT-00059"),
            ("UCS - Sluishoofd - standaard | OBT-00059", "VARIANT - Standaard | OBJ-02452"),
            (
                "VARIANT - Verticaal lierwerk | OBJ-02464",
                "UCS - Bewegingswerk - Sluisdeur - Vertikaal  lierwerk | OBT-00044",
            ),
            (
                "UCS - Bewegingswerk - Sluisdeur - Vertikaal  lierwerk | OBT-00044",
                "VARIANT - Verticaal lierwerk | OBJ-02464",
            ),
            (
                "VARIANT - Vierkantkerend | OBJ-02459",
                "UCS - Sluisdeur - Vierkantkerend | OBT-00023",
            ),
            (
                "UCS - Sluisdeur - Vierkantkerend | OBT-00023",
                "VARIANT - Vierkantkerend | OBJ-02459",
            ),
        ]
    )



"""Tests for the Archimate XML format."""

from pathlib import Path

import pytest
from lxml import etree
from ragraph.datasets import esl
from ragraph.edge import Edge
from ragraph.io import archimate
from ragraph.node import Node


def test_archimate_propdefs(datadir: Path, update: bool):
    g = esl.get("pump")
    elem = archimate._make_property_definitions(g)

    text = etree.tostring(elem, pretty_print=True).decode()
    ref = datadir / "archimate" / "propdefs.xml"
    if update:
        ref.write_text(text)
    assert text == ref.read_text()


def test_archimate_node(datadir: Path, update: bool):
    parent = Node("parent")
    node = Node(
        "testnode",
        parent=parent,
        kind="test",
        labels=["label1", "label2"],
        weights=dict(weight_a=1, weight_b=2),
        annotations=dict(
            annotation_a=1,
            archimate=dict(type="BusinessActor", documentation="Docu with\ntwo strings."),
        ),
    )
    node.is_bus = True
    elem = archimate._node_to_element(node)

    text = etree.tostring(elem, pretty_print=True).decode()
    ref = datadir / "archimate" / "node.xml"
    if update:
        ref.write_text(text)
    assert text == ref.read_text()


def test_archimate_edge(datadir: Path, update: bool):
    edge = Edge(
        name="test",
        source=Node("source"),
        target=Node("target"),
        kind="test",
        labels=["label1", "label2"],
        weights=dict(weight_a=1, weight_b=2),
        annotations=dict(
            annotation_a=1,
            archimate=dict(type="Flow", documentation="Doc with\ntwo lines."),
        ),
    )
    elem = archimate._edge_to_relationship(edge)

    text = etree.tostring(elem, pretty_print=True).decode()
    ref = datadir / "archimate" / "edge.xml"
    if update:
        ref.write_text(text)
    assert text == ref.read_text()


# @pytest.mark.skip(reason="Test fails on UUIDs.")
def test_to_archimate(datadir: Path, update: bool, tmpdir):
    g = esl.get("pump")
    annots = g["world"].annotations
    annots.test_annotation = 12
    annots.archimate = dict(type="CommunicationNetwork", documentation="Docu with\ntwo lines.")

    gr = g["world.drive-mechanism.provide-power"]
    gr.annotations = dict(archimate=dict(type="ApplicationFunction", requirement=gr.name))

    elem = archimate.to_archimate(g, path=Path(tmpdir) / "pump.xml", bundle_schemas=True)

    text = etree.tostring(elem, pretty_print=True).decode()
    ref = datadir / "archimate" / "pump.xml"
    if update:
        ref.write_text(text)
    assert text == ref.read_text()


def test_from_archimate():
    with pytest.raises(NotImplementedError):
        archimate.from_archimate(None)



"""XML format support tests."""

import shutil
from pathlib import Path

import pytest
from lxml import etree
from ragraph import datasets
from ragraph.edge import Edge
from ragraph.generic import Annotations
from ragraph.io import xml
from ragraph.node import Node


def test_xml_annotations():
    # Check conversion.
    a = Annotations(foo=4815162342, bar="foo", baz=-4.815162342, fee=True, empty=None)
    elems = xml._annotations_to_elements(a)
    values = [[(a.tag, a.attrib, a.text) for a in e] for e in elems]
    assert values == [
        [("name", {}, "foo"), ("type", {}, "integer"), ("value", {}, "4815162342")],
        [("name", {}, "bar"), ("type", {}, "string"), ("value", {}, "foo")],
        [("name", {}, "baz"), ("type", {}, "float"), ("value", {}, "-4.815162342")],
        [("name", {}, "fee"), ("type", {}, "boolean"), ("value", {}, "true")],
    ]

    # Check roundtrip.
    rt = xml._elements_to_annotations(elems)
    assert rt == a

    # Check for TypeError on invalid type.
    with pytest.raises(TypeError) as e:
        unsupported = Annotations(a=dict(b=2))
        xml._annotations_to_elements(unsupported)
    assert "Cannot export annotation" in str(e.value)


def test_xml_node():
    parent = Node("parent")
    child = Node("child")
    a = Node(
        "A",
        kind="test",
        parent=parent,
        children=[child],
        is_bus=True,
        labels=["first", "second"],
        weights=dict(a=1, b=2),
        annotations=Annotations(extra="A very nice string\nwith two lines."),
    )

    element_a = xml._node_to_element(a)

    assert element_a.tag == "{http://ratio-case.nl/ragraph}Node"
    assert element_a.attrib == {
        "{http://www.omg.org/spec/XMI/20100901}id": ("_n_00000000-0000-0000-0000-000000000002")
    }

    values = [(e.tag, e.attrib, e.text, [(a.tag, a.attrib, a.text) for a in e]) for e in element_a]
    assert values == [
        ("name", {}, "A", []),
        ("uuid", {}, "00000000-0000-0000-0000-000000000002", []),
        ("kind", {}, "test", []),
        ("label", {}, "first", []),
        ("label", {}, "second", []),
        ("{http://ratio-case.nl/ragraph}Weight", {"value": "1"}, "a", []),
        ("{http://ratio-case.nl/ragraph}Weight", {"value": "2"}, "b", []),
        (
            "{http://ratio-case.nl/ragraph}Annotation",
            {},
            None,
            [
                ("name", {}, "extra"),
                ("type", {}, "string"),
                ("value", {}, "A very nice string\nwith two lines."),
            ],
        ),
        ("parent", {}, "_n_00000000-0000-0000-0000-000000000000", []),
        ("child", {}, "_n_00000000-0000-0000-0000-000000000001", []),
        ("is_bus", {}, "true", []),
    ]

    rt = xml._element_to_node(element_a)
    node_dict = {
        "_n_00000000-0000-0000-0000-000000000002": rt,
        "_n_00000000-0000-0000-0000-000000000000": parent,
        "_n_00000000-0000-0000-0000-000000000001": child,
    }
    xml._resolve_node_references(node_dict)

    assert rt.as_dict(use_uuid=False) == a.as_dict(use_uuid=False)


def test_xml_edge():
    source = Node("source")
    target = Node("target")
    edge = Edge(
        source,
        target,
        name="test",
        kind="test",
        labels=["first", "second"],
        weights=dict(a=1, b=2),
        annotations=Annotations(extra="A very nice string\nwith two lines."),
    )

    elem = xml._edge_to_element(edge)

    assert elem.tag == "{http://ratio-case.nl/ragraph}Edge"

    eid = "_e_00000000-0000-0000-0000-000000000002"
    assert elem.attrib == {"{http://www.omg.org/spec/XMI/20100901}id": eid}

    values = [(e.tag, e.attrib, e.text, [(a.tag, a.attrib, a.text) for a in e]) for e in elem]
    assert values == [
        ("name", {}, "test", []),
        ("uuid", {}, "00000000-0000-0000-0000-000000000002", []),
        ("kind", {}, "test", []),
        ("label", {}, "first", []),
        ("label", {}, "second", []),
        ("{http://ratio-case.nl/ragraph}Weight", {"value": "1"}, "a", []),
        ("{http://ratio-case.nl/ragraph}Weight", {"value": "2"}, "b", []),
        (
            "{http://ratio-case.nl/ragraph}Annotation",
            {},
            None,
            [
                ("name", {}, "extra"),
                ("type", {}, "string"),
                ("value", {}, "A very nice string\nwith two lines."),
            ],
        ),
        ("source", {}, "_n_00000000-0000-0000-0000-000000000000", []),
        ("target", {}, "_n_00000000-0000-0000-0000-000000000001", []),
    ]

    node_dict = {
        "_n_00000000-0000-0000-0000-000000000000": source,
        "_n_00000000-0000-0000-0000-000000000001": target,
    }
    rt = xml._element_to_edge(elem, node_dict)
    rt._source = source
    rt._target = target
    assert rt.as_dict(use_uuid=False) == edge.as_dict(use_uuid=False)


def test_to_xml(tmpdir, datadir, update):
    graph = datasets.get("aircraft_engine")
    path = tmpdir / "aircraft_engine.xml"
    xml.to_xml(
        graph,
        path=path,
        bundle_schemas=False,
    )
    if update:
        shutil.copy2(path, datadir / "xml" / "aircraft_engine.xml")

    p = Path(tmpdir)
    assert p / "XMI-Canonical.xsd" not in p.iterdir()
    assert p / "ragraph.xsd" not in p.iterdir()

    xml.to_xml(graph, path=tmpdir / "test2.xml", bundle_schemas=True)
    assert p / "XMI-Canonical.xsd" in p.iterdir()
    assert p / "ragraph.xsd" in p.iterdir()

    doc = etree.parse(str(xml.here / "bare.xml"))
    xml.to_xml(graph, elem=doc.getroot())

    assert xml.schema.validate(doc)


def test_from_xml(datadir):
    p = datadir / "xml" / "aircraft_engine.xml"
    ref = datasets.get("aircraft_engine")
    ref.name = "aircraft_engine"  # Fix name to be non-unique

    g1 = xml.from_xml(path=p, validate=False)
    assert g1 == ref

    g2 = xml.from_xml(enc=p.read_bytes())
    assert g2 == ref

    g3 = xml.from_xml(elem=etree.parse(str(p)).getroot())
    assert g3 == ref

    with pytest.raises(ValueError) as e:
        xml.from_xml()
    assert "arguments should be set" in str(e.value)

    with pytest.raises(ValueError) as e:
        xml.from_xml(path=datadir / "xml" / "faulty.xml")
    assert "Weiiight" in str(e.value)



import pytest
from ragraph.io.csv import _convert_to_num, _load_edges, from_csv, to_csv


def test_from_csv(datadir):
    # Try from csv in increasing complexity.
    from_csv()  # returns an empty graph
    from_csv(nodes_path=datadir / "csv" / "simple_nodes.csv")
    from_csv(edges_path=datadir / "csv" / "simple_edges.csv")
    from_csv(
        nodes_path=datadir / "csv" / "simple_nodes.csv",
        edges_path=datadir / "csv" / "simple_edges.csv",
    )
    from_csv(nodes_path=datadir / "csv" / "rich_nodes.csv", node_weights=["bank"])
    from_csv(edges_path=datadir / "csv" / "rich_edges.csv", edge_weights=["amount"])

    # Check extensive one.
    g = from_csv(
        nodes_path=datadir / "csv" / "rich_nodes.csv",
        edges_path=datadir / "csv" / "rich_edges.csv",
        node_weights=["bank"],
        edge_weights=["amount"],
    )

    assert g.node_kinds == ["broke", "dagobert", "empty", "poor", "rich", "wealthy"]
    assert g.node_labels == ["default", "labeled", "special"]
    assert g.node_weight_labels == ["bank"]
    assert g.edge_kinds == ["dollars", "euros"]
    assert g.edge_labels == ["eu", "money", "us"]
    assert g.node_dict["e"].is_bus
    assert g.get_hierarchy_dict() == {"a": {"b": {}, "c": {}, "f": {}}, "d": {"e": {}}}
    assert g["a"].annotations.comment == "much nice, very comment"
    assert g["a", "b"][0].annotations.comment == "super useful"


def test_from_csv_errors(datadir):
    with pytest.raises(ValueError) as e:
        from_csv(nodes_path=datadir / "csv" / "wrong.csv")
        assert "at least a 'name' column" in str(e.value)

    with pytest.raises(ValueError) as e:
        from_csv(edges_path=datadir / "csv" / "wrong.csv")
        assert "at least a 'source' and 'target' column" in str(e.value)

    with pytest.raises(ValueError) as e:
        _load_edges(datadir / "csv" / "wrong.csv", ";", ";", dict(), list())
        assert "at least a 'source' and 'target' column" in str(e.value)

    with pytest.raises(ValueError) as e:
        from_csv(nodes_path=datadir / "csv" / "dupe_nodes.csv")
        assert "not unique" in str(e.value)


def test_to_csv(datadir, tmpdir, update):
    g = from_csv(
        nodes_path=datadir / "csv" / "rich_nodes.csv",
        edges_path=datadir / "csv" / "rich_edges.csv",
        node_weights=["bank"],
        edge_weights=["amount"],
    )
    if update:
        to_csv(g, datadir / "csv/rich")
    to_csv(g, tmpdir / "graph")

    nodes_expected = (datadir / "csv" / "rich_nodes.csv").read_text("utf-8")
    nodes_output = (tmpdir / "graph_nodes.csv").read_text("utf-8")

    (datadir / "csv" / "rich_nodes.csv").write_text(nodes_output)

    assert nodes_output == nodes_expected

    edges_expected = (datadir / "csv" / "rich_edges.csv").read_text("utf-8")
    edges_output = (tmpdir / "graph_edges.csv").read_text("utf-8")
    assert edges_output == edges_expected


def test_num_convert():
    assert _convert_to_num(6) == 6, "Integer convert should do nothing."
    assert _convert_to_num(" true") is True, "String bool convert should return True."
    assert _convert_to_num("FALSE  ") is False, "String bool convert should return False."

    with pytest.warns(UserWarning) as w:
        _convert_to_num("faulty")
    assert w[0].message.args[0] == " ".join(
        [
            "You are assigning a string as a weight property.",
            "Please use a float, integer, or a Bool.",
            "A default value of 1.0 is used as a replacement for {}.".format("faulty"),
        ]
    )



"""Tests for datasets module."""

from ragraph import datasets


def test_datasets():
    assert datasets.enum()
    for i in datasets.enum():
        assert datasets.info(i)
        assert datasets.get(i)


def test_symmetry():
    symmetrical = ["climate_control", "localbus", "overlap"]
    for ds in symmetrical:
        g = datasets.get(ds)
        leafs = g.leafs
        dim = len(leafs)

        error = False
        msg = "{}: '{}' weight between '{}' and '{}' does not match."
        for weight in g.edge_weight_labels:
            adj = g.get_adjacency_matrix(nodes=leafs, only=[weight])
            for i in range(dim):
                for j in range(dim):
                    if adj[i][j] != adj[j][i]:
                        print(msg.format(ds, weight, leafs[i].name, leafs[j].name))
                        error = True
        # Fail on the last one. Logged output already shows all.
        assert not error, msg.format(ds, weight, leafs[i].name, leafs[j].name)



from ragraph import datasets
from ragraph.plot import chord


def test_chord(tmpdir, datadir, check_diff):
    g = datasets.get("climate_control")
    fig = chord(g)

    fname = "cc_chord.svg"
    tmp = tmpdir / fname
    fig.save_svg(tmp)
    check_diff(tmp, datadir / "svg" / fname)



"""Test Canopy data format."""

import pytest
from ragraph import datasets
from ragraph.io import canopy


@pytest.fixture(params=["workspace", "session", "tab", "graph"])
def schema(request):
    return request.param


def test_to_canopy(tmp_path):
    g = datasets.get("climate_control_mg")
    canopy.to_canopy(g, tmp_path / "canopy.json")


def test_from_canopy(datadir, schema):
    gs = canopy.from_canopy(datadir / "canopy" / f"{schema}.canopy.json")
    g = gs[0]
    ref = datasets.get("climate_control_mg")
    assert g.name == ref.name
    assert set(n.name for n in g.nodes) == set([n.name for n in ref.nodes])
    assert len(g.edges) == len(ref.edges)



from ragraph import datasets
from ragraph.analysis import heuristics
from ragraph.analysis.heuristics import markov_gamma
from ragraph.graph import Graph


def test_markov_gamma_global_bus(graph: Graph):
    markov_gamma(graph, alpha=2, beta=2.0, mu=2.0, gamma=2.0, local_buses=False)
    assert graph.check_consistency(raise_error=True)
    assert graph.get_hierarchy_dict() == {
        "f": {},
        "node.node0": {"d": {}, "e": {}, "node.node1": {"a": {}, "b": {}, "c": {}}},
    }
    assert [n.name for n in graph.nodes if n.is_bus] == ["e"]


def test_markov_gamma_local_bus(graph: Graph):
    notinplace, _ = markov_gamma(
        graph, alpha=2, beta=2.0, mu=2.0, gamma=2.0, local_buses=True, inplace=False
    )
    markov_gamma(graph, alpha=2, beta=2.0, mu=2.0, gamma=2.0, local_buses=True)

    assert graph.check_consistency(raise_error=True)
    assert graph.get_hierarchy_dict() == {
        "f": {},
        "node.node0": {"d": {}, "e": {}, "node.node1": {"a": {}, "b": {}, "c": {}}},
    }
    assert [n.name for n in graph.nodes if n.is_bus] == ["a", "e"]
    assert graph.get_hierarchy_dict() == notinplace.get_hierarchy_dict()


def test_markov_gamma_reproduceability():
    g = datasets.get("elevator175")

    markov_gamma(
        g,
        root=None,
        leafs=[n.name for n in g.leafs],
        alpha=2,
        beta=2.0,
        mu=2.0,
        gamma=2.0,
        inplace=True,
    )
    res1 = g.get_hierarchy_dict()

    markov_gamma(
        g,
        root=None,
        leafs=[n.name for n in g.leafs],
        alpha=2,
        beta=2.0,
        mu=2.0,
        gamma=2.0,
        inplace=True,
    )
    res2 = g.get_hierarchy_dict()

    assert res1 == res2

    g2, _ = markov_gamma(
        g,
        root=None,
        leafs=[n.name for n in g.leafs],
        alpha=2,
        beta=2.0,
        mu=2.0,
        gamma=2.0,
        inplace=False,
    )
    res3 = g2.get_hierarchy_dict()
    assert res3 == res1


def test_markov_gamma_single_node(graph: Graph):
    assert markov_gamma(graph, leafs=["a"], names=True)[1] == ["a"]


def test_markov_gamma_cases(case: Graph):
    heuristics.markov_gamma(case)



"""Tests for the Mapping object used in the plot module."""

import pytest
from ragraph.generic import Mapping, MappingValidationError


def test_mapping_basics():
    class SimpleMap(Mapping):
        _protected = False

    # Test instantiation
    m = SimpleMap(a=1, c=SimpleMap(bla="bla"))
    assert m.get("_data") == m._data
    assert m.a == 1

    # Update
    m.update(SimpleMap(a=2, b="2", c=SimpleMap(foo="bar")))

    # Check the getting methods
    assert m.get("a") == 2
    assert m.a == 2
    assert m["a"] == 2
    assert m.get("b") == "2"
    assert m.b == "2"
    assert m["b"] == "2"

    # Check updating a submap
    m["c"] = dict(foo="qux")
    assert m.c.as_dict() == dict(bla="bla", foo="qux")

    # Some other datatypes
    m.d = {"baz"}
    assert m["d"] == {"baz"}
    m["e"] = []
    assert m.get("e") == []

    # Check iterator
    for k, v in m:
        assert m.get(k) == v

    # Check final dict representation
    assert m.as_dict() == {
        "a": 2,
        "b": "2",
        "c": {"bla": "bla", "foo": "qux"},
        "d": {"baz"},
        "e": [],
    }

    # Validation should not do anything, but certainly not fail.
    m.validate()


def test_mapping_validation():
    def assert_int(x):
        assert isinstance(x, int)

    class SimpleMap(Mapping):
        _protected = False
        _validators = dict(foo=assert_int, bar=assert_int)

    m = SimpleMap()
    m.foo = 2
    m.bar = "2"

    assert m._data["foo"] == 2
    assert str(m) == "SimpleMap({'foo': 2, 'bar': '2'})"

    # Check validation.
    with pytest.raises(MappingValidationError) as e:
        m.validate()
    assert "validation of 'bar' failed" in str(e).lower()


def test_mapping_post_validation():
    msg = "If you never fail, you will never succeed."

    class SimpleMap(Mapping):
        def _post_validation():
            raise Exception(msg)

    m = SimpleMap()

    with pytest.raises(MappingValidationError) as e:
        m.validate()
    assert "Post-validation did not succeed" in str(e)


def test_mapping_update_error():
    m = Mapping()

    with pytest.raises(TypeError) as e:
        m.update(6)
    assert "Expected a dictionary or a Mapping" in str(e)



"""Tests for node.py."""

import pytest

from ragraph.node import Node


@pytest.fixture
def rich_node():
    return Node(
        "rich",
        kind="wealthy",
        parent=Node("parent"),
        children=[Node("a"), Node("b", children=[Node("c")])],
        weights=dict(dollars=1, euros=0.5),
        annotations=dict(foo="bar"),
    )


def test_node_properties(rich_node):
    assert str(rich_node), "Should be able to get str()."
    assert repr(rich_node), "Should be able to get repr()."
    assert rich_node.name == "rich"
    assert rich_node.kind == "wealthy"
    assert rich_node.parent.name == "parent"
    assert [c.name for c in rich_node.children] == ["a", "b"]
    assert rich_node.weights == dict(dollars=1, euros=0.5)
    assert rich_node.weight == 1.5
    assert rich_node.annotations.foo == "bar"
    assert rich_node.ancestors == [rich_node.parent]
    assert rich_node.descendants == rich_node.children + rich_node.children[1].children
    assert rich_node.depth == 1
    assert rich_node.height == 2
    assert rich_node.width == 2
    assert not rich_node.is_leaf
    assert not rich_node.is_root

    rich_node.parent = None
    assert str(rich_node), "Should be able to get str() with empty parent."


def test_node_instantiations():
    assert Node(6).name == "6"

    assert Node("test", kind=None).kind == "default", "Node should have 'default' kind."

    assert (
        Node("test", annotations=dict(a=1)).annotations.a == 1
    ), "Node annotations should accept a dict as argument."

    with pytest.raises(TypeError) as e:
        Node("test", parent=5)
    assert str(e.value) == "Parent should be of type {}, got {}.".format(Node, int)

    with pytest.raises(ValueError) as e:
        Node("test", is_bus=True)
    assert str(e.value) == "Cannot be a bus without a parent to be it for. Set a parent first!"


def test_node_representation(rich_node):
    assert str(rich_node) == "Node(name='rich')"
    non_id = repr(rich_node).split(", uuid=")[0]
    ref = (
        "<ragraph.node.Node(name='rich', "
        + "parent='parent', children=[\"Node(name='a')\", \"Node(name='b')\"], "
        + "is_bus=False, kind='wealthy', labels=['default'], "
        + "weights={'dollars': 1, 'euros': 0.5}, "
        + "annotations=Annotations({'foo': 'bar'})"
    )
    assert non_id == ref


def test_node_hierarchy_changes(a, b, c, d, e, f):
    a.children = [b, c]
    assert a.children == [b, c]
    assert b.parent == a
    assert c.parent == a

    # Different way of adding children
    d.children = [e]
    f.parent = d
    assert d.children == [e, f]
    assert e.parent == d
    assert f.parent == d

    # Resetting parent of f
    f.parent = None
    assert d.children == [e]
    assert f.parent is None

    # Move children of a to f
    f.children = a.children
    assert b.parent == f
    assert c.parent == f
    assert a.children == []

    # Changing parent of b back to a
    b.parent = a
    assert f.children == [c]
    assert c.parent == f
    assert a.children == [b]
    assert b.parent == a
    assert b.parent == a



"""Tests for _classes module."""

import pytest
from ragraph import datasets
from ragraph.analysis import _classes
from ragraph.node import Node


def test_cast_abstract():
    with pytest.raises(Exception):
        _classes.Cast()  # Cannot instantiate an abstract class.

    class Dummy(_classes.Cast):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)

        def __call__(self, value):
            super().__call__(value)

    d = Dummy()
    assert d(None) is None


def test_node_cast():
    graph = datasets.get("climate_control")
    nc = _classes.NodeCast(graph)
    node = nc("Accumulator")
    assert isinstance(node, Node)

    ref = graph.node_dict["Radiator"]
    assert nc(ref) == ref

    with pytest.raises(ValueError) as e:
        nc.graph = None
        nc("Accumulator")
    assert "must be set" in str(e.value)


def test_parameter_numeric():
    p = _classes.Parameter(
        "test",
        float,
        description="descr.",
        default=True,
        cast=float,
        lower=_classes.Bound(0.0, report="error"),
        upper=_classes.Bound(3.0, report="error"),
    )
    assert p.default == 1.0 and isinstance(p.default, float)
    assert p.parse(2) == 2.0

    with pytest.raises(ValueError) as e:
        p.parse(-1.0)
    assert "below" in str(e.value)

    with pytest.raises(ValueError) as e:
        p.parse(3.1)
    assert "above" in str(e.value)


def test_parameter_enum():
    opts = ["yes", 1, "maybe"]
    p = _classes.Parameter("test", str, description="descr.", default=True, cast=str, enum=opts)
    assert p.enum == {"yes", "1", "maybe"}

    p.parse("yes")

    with pytest.raises(ValueError) as e:
        p.parse("nope")
    assert "not allowed" in str(e.value)



"""# Chord plots usings openchord"""

from typing import Any, Dict, List, Optional, Union

from ragraph.graph import Graph
from ragraph.node import Node
from ragraph.plot.generic import Style

try:
    import rachord
except ImportError:
    raise ModuleNotFoundError(
        "Could not find the `rachord` dependency. To use the chord plot functionality, "
        "RaGraph needs to be installed as `ragraph[plot]` or `ragraph[all]`."
    )


def chord(
    graph: Graph,
    nodes: Optional[Union[List[Node], List[str]]] = None,
    style: Optional[Style] = None,
    adj_kwargs: Optional[Dict[str, Any]] = None,
    symmetrize: bool = False,
    show: bool = False,
) -> rachord.Chord:
    """Make a chord plot for the given nodes in a graph.

    Arguments:
        graph: Graph to create a chord plot for.
        nodes: Nodes or node names to include in the chord plot. Defaults to leaf nodes.
        style: Plotting style.
        adj_kwargs: Additional arguments to [`ragraph.graph.Graph.get_adjacency_matrix`
            ][ragraph.graph.Graph.get_adjacency_matrix].
        symmetrize: Whether to symmetrize the adjacency matrix.
        show: Whether to show the resulting figure.

    Returns:
        Chord plot using rachord.
    """
    nodes = graph.leafs if nodes is None else nodes
    nodes = [n if isinstance(n, Node) else graph.node_dict[n] for n in nodes]
    labels = [n.name for n in nodes]

    adj_kwargs = dict() if adj_kwargs is None else adj_kwargs
    adj = graph.get_adjacency_matrix(nodes=nodes, **adj_kwargs)

    if symmetrize:
        dim = len(nodes)
        for row in range(dim):
            for col in range(row, dim):
                value = (adj[row][col] + adj[col][row]) / 2
                adj[row][col] = value
                adj[col][row] = value

    style: Style = Style() if style is None else style
    cs = style.chord
    fig = rachord.Chord(adj, labels, radius=cs.radius)
    fig.padding = (
        cs.fontsize * cs.fontfactor * max(len(x) for x in labels)
        if cs.padding is None
        else cs.padding
    )
    fig.gap_size = cs.gap_size
    fig.ribbon_gap = cs.ribbon_gap
    fig.ribbon_stiffness = cs.ribbon_stiffness
    fig.arc_thickness = cs.arc_thickness
    fig.bg_color = cs.bg_color
    fig.bg_transparency = cs.bg_transparency
    fig.font_size = cs.fontsize
    fig.font_family = cs.fontfamily

    fig.colormap = style.palettes.categorical

    if show:
        return fig.show()
    else:
        return fig



"""Tests for compatibility analysis module."""

from collections import defaultdict
from pathlib import Path

import numpy as np
import pytest
from ragraph import datasets
from ragraph.analysis import compatibility
from ragraph.graph import Graph


@pytest.fixture
def compat():
    return datasets.get("compatibility").get_adjacency_matrix()


@pytest.fixture
def compat_graph():
    return datasets.get("compatibility")


def test_feasible_configs(compat):
    assert compatibility.is_feasible(compat, (1, 4)), "Should be feasible."
    # [1,2,2] means: always pick 0, pick 1 or 2, pick 3 or 4
    # in this example, we skip 5.
    assert list(compatibility.yield_feasible_configurations(compat, [1, 2, 2])) == [
        (0, 2, 4),
        (0, 1, 4),
    ]


def test_compatibility_analysis(compat_graph: Graph, compat: np.ndarray, datadir: Path, tmpdir):
    variants = defaultdict(list)
    for n in compat_graph.nodes:
        variants[n.kind].append(n)
    ca = compatibility.CompatibilityAnalysis(
        compat_graph,
        variants=variants,
    )
    comps = ca._variants_list
    assert [n.name for n in comps] == ["A1", "B1", "B2", "C1", "C2", "C3"]

    assert np.all(np.array(ca.get_compatibility_matrix()) == np.array(compat))

    assert list(ca.yield_feasible_configurations()) == [
        tuple(comps[i] for i in (0, 2, 5)),
        tuple(comps[i] for i in (0, 2, 4)),
        tuple(comps[i] for i in (0, 1, 5)),
        tuple(comps[i] for i in (0, 1, 4)),
    ], "Feasible configs in terms of nodes."

    assert ca.get_ranked_configurations() == [
        (tuple(comps[i] for i in (0, 2, 5)), 6.0),
        (tuple(comps[i] for i in (0, 2, 4)), 5.0),
        (tuple(comps[i] for i in (0, 1, 5)), 5.0),
        (tuple(comps[i] for i in (0, 1, 4)), 4.0),
    ], "Ranked configs."

    ca.score_method = compatibility.get_score_method(variant_agg="product", config_agg="product")
    assert ca.get_config_score(tuple(comps[i] for i in (0, 2, 4))) == 4.0

    ca.write_csv(datadir / "csv" / "feasible.csv")
    assert Path(datadir / "csv" / "feasible.csv").read_text().split("\n") == [
        "A;B;C;score",
        "A1;B2;C3;6.0",
        "A1;B2;C2;4.0",
        "A1;B1;C3;3.0",
        "A1;B1;C2;2.0",
        "",
    ]

    fig = ca.plot()
    fig.write_image(tmpdir / "ca.svg", format="svg")



"""Tests for the plot generic module."""

from plotly import graph_objs as go
from ragraph.plot import generic


def test_treestyle():
    foo = generic.TreeStyle(xaxis=dict(showgrid=True), yaxis=dict(zeroline=True))
    assert isinstance(foo.xaxis, go.layout.XAxis)

    assert foo.xaxis.showgrid
    assert foo.yaxis.zeroline

    assert not foo.xaxis.showline
    assert not foo.yaxis.showline

    assert foo.line.color == "gray"


def test_labelsstyle():
    foo = generic.LabelsStyle(
        xaxis=dict(automargin=True), yaxis=dict(autorange=False), fontfamily="Arial"
    )

    assert foo.xaxis.automargin
    assert foo.yaxis == foo._defaults["yaxis"]

    assert foo.fontfamily == "Arial"


def test_piemapstyle():
    foo = generic.PieMapStyle(
        xaxis=dict(automargin=True), yaxis=dict(scaleratio=0.5), mode="relative"
    )

    assert foo.xaxis.automargin
    assert foo.yaxis.scaleratio == 0.5
    assert foo.mode == "relative"
    assert foo.display == "kinds"


def test_legendstyle():
    foo = generic.LegendStyle(height=7, n_ticks=3)
    assert foo.height == 7
    assert foo.n_ticks == 3


def test_palettes():
    foo = generic.Palettes()
    assert len(foo.categorical) == 10
    assert len(foo.continuous) == 256
    assert isinstance(foo.fields, dict)
    assert isinstance(foo.domains, dict)


def test_style():
    from ragraph.generic import Mapping

    foo = generic.Style()

    assert isinstance(foo.piemap, Mapping)


def test_component():
    from plotly import graph_objs as go

    foo = generic.Component(
        xaxis=dict(zeroline=False), yaxis=dict(automargin=False), width=10, height=10
    )

    assert not foo.xaxis.zeroline
    assert not foo.yaxis.automargin

    assert isinstance(foo.get_figure(show=False), go.Figure)



"""Tests for the SVG class."""

from ragraph.plot import svg


def test_line():
    line = svg.Line()

    assert line.width == 2


def test_svg():
    s = svg.SVG()
    assert s.as_dict() == dict()


def test_gets():
    assert svg.get_line(x0=0.0, x1=1.0, y0=2.0, y1=3.0).as_dict() == dict(
        type="line", x0=0.0, x1=1.0, y0=2.0, y1=3.0
    )

    assert svg.get_curvedline(x0=0.0, x1=1.0, x2=2.0, y0=3.0, y1=4.0, y2=5.0).as_dict() == dict(
        type="path", path="M 0.0 3.0 Q 1.0 4.0 2.0 5.0"
    )

    assert svg.get_rectangle(x0=0.0, x1=1.0, y0=2.0, y1=3.0).as_dict() == dict(
        type="rect", x0=0.0, x1=1.0, y0=2.0, y1=3.0
    )

    assert svg.get_wedge(
        x=-1.0, y=0.0, r=1.0, start_angle=0.0, end_angle=svg.pi / 3
    ).as_dict() == dict(
        type="path",
        path="M -1.0 0.0 L 0.0 0.0 Q 0.0 0.414 -0.293 0.707 Q -0.386 0.8 -0.5 0.866 Z",
    )



from ragraph.graph import Graph
from ragraph.io.esl import from_esl


def test_from_esl(esl_file):
    g: Graph = from_esl(esl_file)

    assert g.node_count == 8, "Graph of test.esl should have 5 nodes, not %d." % g.node_count
    assert g.edge_count == 1, "Graph of test.esl should have 9 edges, not %d." % g.edge_count

    g.check_consistency()



from pathlib import Path

from pytest import mark
from ragraph import datasets
from ragraph.analysis.heuristics import markov_gamma
from ragraph.colors import get_cyan, get_green, get_purple, get_red
from ragraph.edge import Edge
from ragraph.node import Node
from ragraph.plot import mdm
from ragraph.plot.generic import Style

SHOW_FIGURES = True


@mark.skip(reason="Dict comparison fails randomly.")
def test_mdm_0(check_plotly):
    g = datasets.get("climate_control")

    # Categorical plot of weight labels with custom colors
    fig = mdm(
        leafs=g.leafs,
        edges=g.edges,
        style=Style(
            piemap={
                "display": "weight labels",
                "mode": "equal",
            },
            palettes={
                "fields": {
                    "spatial": "#d21404",
                    "energy flow": "#d7be69",
                    "information flow": "#ffa500",
                    "material flow": "#ffa500",
                }
            },
        ),
        sort=False,
    )

    fname = "categorical_plot_with_custom_colors.json"
    check_plotly(fig, fname)


@mark.skip(reason="Dict comparison fails randomly.")
def test_mdm_1(check_plotly):
    g = datasets.get("climate_control")

    # Numerical plot of weights with custom colors.
    fig = mdm(
        leafs=g.leafs,
        edges=g.edges,
        style=Style(
            piemap={
                "display": "weights",
                "mode": "equal",
            },
            palettes={
                "fields": {
                    "spatial": get_red(),
                    "energy flow": get_green(),
                    "information flow": get_cyan(),
                    "material flow": get_purple(),
                }
            },
        ),
        sort=False,
    )

    fname = "numerical_plot_with_custom_colors.json"
    check_plotly(fig, fname)


@mark.skip(reason="Dict comparision fails randomly.")
def test_mdm_2(check_plotly):
    g = datasets.get("climate_control")

    # Plot clustered matrix
    markov_gamma(g, alpha=2, beta=2.0, mu=2.0, gamma=2.0, local_buses=True)
    fig = mdm(
        leafs=g.leafs,
        edges=g.edges,
        style=Style(
            piemap={
                "display": "weight labels",
                "mode": "equal",
                "busarea": {"fillcolor": "#d21404"},
            },
        ),
        sort=True,
    )

    fname = "clustered_matrix_plot.json"
    check_plotly(fig, fname)


@mark.skip(reason="Dict comparision fails randomly.")
def test_mdm_3(check_plotly):
    # Plot matrix with more than 10 edge labels.
    nodes = [Node("A", kind="component"), Node("B", kind="component")]
    edges = [
        Edge(
            source=nodes[0],
            target=nodes[1],
            kind="functional_dependency",
            labels=[str(idx) for idx in range(12)],
        )
    ]

    fig = mdm(
        leafs=nodes,
        edges=edges,
        style=Style(piemap={"display": "labels", "mode": "relative"}),
        sort=False,
    )

    fname = "graph_with_more_than_10_edge_labels.json"
    check_plotly(fig, fname)


@mark.skip(reason="Dict comparision fails randomly.")
def test_mdm_4(check_plotly):
    g = datasets.get("aircraft_engine")

    # Plot clustered matrix
    markov_gamma(g, alpha=2, beta=2.0, mu=2.0, gamma=2.0, local_buses=True)
    fig = mdm(
        leafs=g.leafs,
        edges=g.edges,
        style=Style(
            piemap={
                "display": "weight labels",
                "mode": "equal",
            },
        ),
        sort=True,
    )

    fname = "aircraft_engine.json"
    check_plotly(fig, fname)


@mark.skip(reason="Dict comparision fails randomly.")
def test_mdm_5(check_plotly):
    g = datasets.get("climate_control")

    fig = mdm(
        leafs=g.leafs,
        edges=g.edges,
        style=Style(
            piemap={
                "display": "weight labels",
                "mode": "equal",
            },
            palettes={
                "fields": {
                    "spatial": "#d21404",
                    "energy flow": "#d7be69",
                    "information flow": "#ffa500",
                    "material flow": "#ffa500",
                }
            },
            row_col_numbers=False,
        ),
        sort=False,
    )

    fname = "mdm_no_row_col_numbers.json"
    check_plotly(fig, fname)


@mark.skip(reason="Dict comparision fails randomly.")
def test_mdm_6(check_plotly):
    from ragraph.generic import Annotations

    g = datasets.get("climate_control")

    num = [idx for idx in range(len(g.edges))]

    for idx, e in enumerate(g.edges):
        e.annotations = Annotations(blab=num[idx], blib=num[idx])

    fig = mdm(
        leafs=g.leafs,
        edges=g.edges,
        style=Style(
            piemap={
                "customhoverkeys": ["blib", "blab"],
            },
        ),
        sort=False,
    )

    fname = "mdm_custom_hover.json"
    check_plotly(fig, fname)


def test_highlight(tmpdir):
    g = datasets.get("climate_control_mg")
    g["Accumulator"].annotations.highlight = True
    fig = mdm(g.leafs, g.edges)
    fig.write_image(Path(tmpdir / "highlight.png"), format="png")

    g["Heater Core"].annotations.highlight = "red"
    fig = mdm(g.leafs, g.edges)
    fig.write_image(Path(tmpdir / "custom_highlight.png"), format="png")



from collections import defaultdict
from pathlib import Path

import numpy as np
import pytest
from ragraph.analysis import compatibility as comp
from ragraph.graph import Edge, Graph, Node
from ragraph.io import canopy, csv

EXPORT_GRAPH = False
EXPORT_FIG = False


@pytest.mark.skipif(not EXPORT_GRAPH, reason="No graph export.")
def test_make_rws_graph(datadir):
    import pandas as pd

    r = datadir / "rws"
    mod_fun_con_emb = pd.read_csv(r / "modules-functions-concepts-embodiments.csv", sep=";")
    con_incompat = pd.read_csv(r / "concept-incompatibility-interfaces.csv", sep=";")
    obj_domains = pd.read_csv(r / "objects-domains.csv", sep=";")
    objs = pd.read_csv(r / "objects.csv", sep=";")
    func_if = pd.read_csv(r / "function-interfaces.csv", sep=";")
    domains = pd.read_csv(r / "domains.csv", sep=";")
    con_domains = pd.read_csv(r / "concepts-domains.csv", sep=";")

    obj_nodes = {
        id: Node(name=id, kind="object", annotations=dict(display_name=name))
        for _, id, name in objs.itertuples()
    }

    dom_nodes = {
        name: Node(name=name, kind="constraint")
        for name in set(domains.Domain.unique()).union(domains.Subdomain.unique())
    }
    for _, dom, subdom in domains.itertuples():
        dom_nodes[subdom].parent = dom_nodes[dom]

    elem_nodes = {}
    con_name_to_id = {}
    for (
        _,
        mid,
        mname,
        fid,
        fname,
        cid,
        cname,
        eid,
        ename,
    ) in mod_fun_con_emb.itertuples():
        if mid not in elem_nodes:
            elem_nodes[mid] = Node(
                name=mid,
                kind="element",
                labels=["module"],
                annotations=dict(display_name=mname),
            )
        if fid not in elem_nodes:
            elem_nodes[fid] = Node(
                name=fid,
                kind="element",
                labels=["function"],
                parent=elem_nodes[mid],
                annotations=dict(display_name=fname),
            )
        if cid not in elem_nodes:
            con_name_to_id[cname] = cid
            elem_nodes[cid] = Node(
                name=cid,
                kind="element",
                labels=["concept"],
                parent=elem_nodes[fid],
                annotations=dict(display_name=cname),
            )
        if eid not in elem_nodes:
            elem_nodes[eid] = Node(
                name=eid,
                kind="element",
                labels=["embodiment"],
                parent=elem_nodes[cid],
                annotations=dict(display_name=ename),
            )

    graph = Graph(
        nodes=list(obj_nodes.values()) + list(dom_nodes.values()) + list(elem_nodes.values())
    )

    doms_by_obj = obj_domains.set_index("Object code").drop("Name", axis=1)
    doms_by_con = con_domains.set_index("Concepts")

    for dom_node in dom_nodes.values():
        if dom_node.name in doms_by_obj:
            dom_col = doms_by_obj[dom_node.name].notna()
            for obj_name in dom_col.index[dom_col]:
                obj_node = obj_nodes[obj_name]
                graph.add_edge(
                    Edge(
                        name=f"scope_{dom_node.name}_{obj_name}",
                        source=dom_node,
                        target=obj_node,
                        kind="scope",
                    )
                )
                graph.add_edge(
                    Edge(
                        name=f"scope_{obj_name}_{dom_node.name}",
                        source=obj_node,
                        target=dom_node,
                        kind="scope",
                    )
                )

        if dom_node.name in doms_by_con:
            dom_col = doms_by_con[dom_node.name].notna()
            for con_name in dom_col.index[dom_col]:
                con_node = elem_nodes[con_name_to_id[con_name]]
                graph.add_edge(
                    Edge(
                        name=f"applicability_{dom_node.name}_{con_name}",
                        source=dom_node,
                        target=con_node,
                        kind="applicability",
                    )
                )
                graph.add_edge(
                    Edge(
                        name=f"applicability_{con_name}_{dom_node.name}",
                        source=con_node,
                        target=dom_node,
                        kind="applicability",
                    )
                )

    for _, f_a, fn_a, _, f_b, fn_b, _ in func_if.itertuples():
        a = elem_nodes[f_a]
        b = elem_nodes[f_b]
        assert a.annotations.display_name == fn_a
        assert b.annotations.display_name == fn_b
        try:
            graph.add_edge(
                Edge(
                    name=f"interface_{a.name}_{b.name}",
                    source=a,
                    target=b,
                    kind="interface",
                )
            )
            graph.add_edge(
                Edge(
                    name=f"interface_{b.name}_{a.name}",
                    source=b,
                    target=a,
                    kind="interface",
                )
            )
        except ValueError:
            pass

    for row, c_a, cn_a, _, c_b, cn_b in con_incompat.itertuples():
        a = elem_nodes[c_a]
        b = elem_nodes[c_b]
        assert a.annotations.display_name == cn_a, f"{row}: {a.annotations.display_name} != {cn_a}"
        assert b.annotations.display_name == cn_b, f"{row}: {b.annotations.display_name} != {cn_b}"
        try:
            graph.add_edge(
                Edge(
                    name=f"incompatibility_{a.name}_{b.name}",
                    source=a,
                    target=b,
                    kind="incompatibility",
                )
            )
            graph.add_edge(
                Edge(
                    name=f"incompatibility_{b.name}_{a.name}",
                    source=b,
                    target=a,
                    kind="incompatibility",
                )
            )
        except ValueError:
            pass

    csv.to_csv(graph, r / "config_graph")
    canopy.to_canopy(graph, r / "config_graph.json", fmt="graph")


@pytest.fixture
def rws_graph(datadir: Path):
    r = datadir / "rws"
    return csv.from_csv(r / "config_graph_nodes.csv", r / "config_graph_edges.csv")


def test_rws(rws_graph: Graph, tmpdir):
    variants = defaultdict(list)
    for n in rws_graph.nodes:
        if n.kind == "element" and "concept" in n.labels:
            assert n.parent is not None
            variants[n.parent.name].append(n)

    ca = comp.CompatibilityAnalysis(
        rws_graph,
        variants=variants,
        compatibility_method=comp.get_compatibility_method(
            compatibility_kind=None, incompatibility_kind="incompatibility"
        ),
    )
    assert len(ca._variants_list) == 132
    assert len(ca.variants) == 70

    mat = ca.get_compatibility_matrix()
    assert isinstance(mat, np.ndarray)
    assert np.sum(mat) == 16506.0

    comp.HAVE_NUMPY = False
    mat = ca.get_compatibility_matrix()
    assert isinstance(mat, list)

    if EXPORT_FIG:
        fig = ca.plot()
        fig.write_image(tmpdir / "cim.svg", format="svg")


def test_rws_locks(rws_graph: Graph, datadir: Path):
    variants = defaultdict(list)
    for n in rws_graph.nodes:
        if n.kind == "element" and "concept" in n.labels:
            if not n.parent:
                raise ValueError("Should have a parent.")
            variants[n.parent.name].append(n)

    for obj in [n for n in rws_graph.nodes if n.kind == "object"]:
        constraints = [n for n in rws_graph.targets_of(obj) if n.kind == "constraint"]

        ca = comp.CompatibilityAnalysis(
            rws_graph,
            variants=variants,
            constraints=constraints,
            compatibility_method=comp.get_compatibility_method(
                compatibility_kind=None, incompatibility_kind="incompatibility"
            ),
        )
        ca.write_csv(
            datadir / "rws" / f"{obj.name}_{obj.annotations['display_name']}_feasible.csv",
            scored=False,
        )

    assert ca.disabled_elements == [
        "f17",
        "f18",
        "f24",
        "f25",
        "f31",
        "f32",
        "f35",
        "f38",
    ]


def test_rws_plot(rws_graph: Graph, tmpdir: Path):
    variants = defaultdict(list)
    for n in rws_graph.nodes:
        if n.kind == "element" and "concept" in n.labels:
            assert n.parent is not None
            variants[n.parent.name].append(n)

    ca = comp.CompatibilityAnalysis(
        rws_graph,
        variants=variants,
        interface_method=comp.get_interface_method(),
        compatibility_method=comp.get_compatibility_method(
            compatibility_kind=None, incompatibility_kind="incompatibility"
        ),
    )
    graph = ca.get_plot_graph()
    assert set(graph.node_labels) == {"function", "concept", "module", "embodiment"}
    if EXPORT_FIG:
        fig = ca.plot()
        fig.write_image(tmpdir / "cim.svg", format="svg")



from pathlib import Path

import pytest
from pytest_examples import CodeExample, EvalExample, find_examples

DOCS_DIR = Path(__file__).parent.parent / "docs"
DOCS_FILES = DOCS_DIR.glob("**/*.md")


@pytest.mark.parametrize("example", find_examples(*DOCS_FILES), ids=str)
def test_docs(example: CodeExample, eval_example: EvalExample):
    """Test the package's documentation."""
    eval_example.set_config(line_length=88)
    if eval_example.update_examples:
        eval_example.format(example)
        eval_example.run_print_update(example)
    else:
        eval_example.lint(example)
        eval_example.run_print_check(example)



"""Jaccard Similarity Index tests."""

import numpy as np
import pytest
from ragraph.analysis import similarity
from ragraph.analysis.similarity._jaccard import _calculate, mapping_matrix
from ragraph.analysis.similarity.utils import (
    on_checks,
    on_contains,
    on_hasattrs,
    on_hasweights,
)


@pytest.fixture
def objects():
    return [{"a", 0, "c", "d", 1}, {0, 1, "c", "d", "e"}, {0, "b", "c", "d", "e"}]


@pytest.fixture
def contents():
    return ["a", "b", "c", "d", "e"]


def test_jaccard_calculate():
    a1 = np.array([True, False, True, True, False])
    a2 = np.array([False, False, True, True, True])
    assert _calculate(a1, a2) == 0.5, "Similarity should be 2.0/4.0 == 0.5."
    assert _calculate(np.array([]), np.array([])) == 0.0


def test_on_contents(objects, contents):
    on = on_contains(contents)
    assert on(objects[0]) == [True, False, True, True, False]
    assert on(objects[1]) == [False, False, True, True, True]
    assert on(objects[2]) == [False, True, True, True, True]


def test_on_checks(objects):
    checks = [lambda obj: 0 in obj, lambda obj: 1 in obj, lambda obj: -1 in obj]
    on = on_checks(checks)
    assert on(objects[0]) == [True, True, False]
    assert on(objects[1]) == [True, True, False]
    assert on(objects[2]) == [True, False, False]


def test_on_hasattrs(objects, contents):
    from ragraph.generic import Annotations

    # Only convert strings to key:key values
    dicts = [{k: k for k in obj if isinstance(k, str)} for obj in objects]
    # Convert to Annotations to obtain objects with attributes
    annots = [Annotations(**d) for d in dicts]

    on = on_hasattrs(contents)
    assert on(annots[0]) == [True, False, True, True, False]
    assert on(annots[1]) == [False, False, True, True, True]
    assert on(annots[2]) == [False, True, True, True, True]


def test_on_hasweights():
    from ragraph.node import Node

    nodes = [
        Node("a", weights=dict(a=2, b=4, d=1)),
        Node("b", weights=dict(a=1, b=6, d=2)),
        Node("c", weights=dict(a=1, b=4, c=6, d=1)),
    ]

    on = on_hasweights(["a", "b", "c", "d"], threshold=2.0)
    ref = np.array([[1, 1, 0, 0], [0, 1, 0, 1], [0, 1, 1, 0]])

    assert (mapping_matrix(nodes, on) == ref).all()


def test_mapping_matrix(objects, contents):
    mapping = mapping_matrix(objects, on_contains(contents))
    ref = np.array(
        [
            [True, False, True, True, False],
            [False, False, True, True, True],
            [False, True, True, True, True],
        ]
    )
    assert (mapping == ref).all()


def test_jaccard_index(objects, contents):
    a, b = objects[0], objects[1]
    assert similarity.jaccard_index(a, b, on_contains(["c", "d"])) == 1.0
    assert similarity.jaccard_index(a, b, on_contains(contents)) == 0.5


def test_jaccard_matrix(objects, contents):
    matrix = similarity.jaccard_matrix(objects, on_contains(contents))
    ref = np.array([[1.0, 0.5, 0.4], [0.5, 1.0, 0.75], [0.4, 0.75, 1.0]])
    assert (matrix == matrix.T).all()
    assert (matrix == ref).all()
